use bytesize::ByteSize;
use clap::Parser;
use regex::Regex;
use reqwest::{header, Client as HTTPClient};
use serde::Deserialize;
use std::fs::{read_to_string, write};
use std::str::from_utf8;
use tokio::time::{sleep, Duration};
#[cfg(feature = "sort")]
#[path = "./sort/main.rs"]
mod sort;
#[cfg(feature = "sort")]
use sort::sort_and_move;

const CONFIG_FILE_NAME: &str = "ab_smalls.toml";
const LONG_ABOUT: &str =
    "ab-smalls helps you download small torrents so you can get good yen yield with limited storage.
DO NOT PASS YOUR PASSKEY AS AN ARG IF YOU ARE ON A SHARED BOX.
command line args are not guaranteed to be hidden from other users so setup a toml file config instead.";

#[derive(Parser, Deserialize, Default)]
#[command(author, version, about, long_about = LONG_ABOUT)]
struct Config {
    /// passkey from ab, grab from your user settings
    #[arg(short, long)]
    passkey: Option<String>,
    /// ab username, case sensitive
    #[arg(short, long)]
    username: Option<String>,
    /// page to start from (inclusive)
    #[arg(short, long)]
    starting_page: Option<u16>,
    /// page to end on (inclusive), default is starting page + 5
    #[arg(short, long)]
    ending_page: Option<u16>,
    /// which type to search, anime or music
    #[arg(short = 't', long, default_value = "anime")]
    search_type: Option<String>,
    /// max size for individual torrents, supports human readable inputs like 100 MiB or 5GB
    #[arg(short, long)]
    max_size: Option<String>,
}

impl Config {
    fn merge(self, other: Config) -> Self {
        let starting_page = self.starting_page.or(other.starting_page).or(Some(1));
        let ending_page = Some(starting_page.unwrap() + 5);
        Self {
            passkey: self.passkey.or(other.passkey),
            username: self.username.or(other.username),
            starting_page: starting_page,
            ending_page: self.ending_page.or(other.ending_page).or(ending_page),
            search_type: self.search_type.or(other.search_type),
            max_size: self.max_size.or(other.max_size),
        }
    }
}

#[derive(Deserialize)]
struct ABError {
    error: String,
}

#[derive(Deserialize)]
struct ScrapeResult {
    #[serde(rename(deserialize = "Groups"))]
    groups: Vec<Group>,
}

#[derive(Deserialize)]
struct Group {
    #[serde(rename(deserialize = "ID"))]
    id: u32,
    #[serde(rename(deserialize = "Torrents"))]
    torrents: Vec<Torrent>,
}

#[derive(Deserialize)]
struct Torrent {
    #[serde(rename(deserialize = "ID"))]
    id: u32,
    #[serde(rename(deserialize = "Link"))]
    link: String,
    #[serde(rename(deserialize = "Size"))]
    size: u64,
}

#[tokio::main]
async fn main() {
    let contents = match read_to_string(CONFIG_FILE_NAME) {
        Ok(c) => c,
        Err(_) => {
            println!(
                "[WARN] could not find/read config file `{}`, skipping",
                CONFIG_FILE_NAME
            );
            "".to_string()
        }
    };
    let args: Config = match toml::from_str(&contents) {
        Ok(t) => t,
        Err(err) => {
            println!(
                "[WARN] failed to parse toml file, using defaults err={}",
                err
            );
            Default::default()
        }
    };
    let args = args.merge(Config::parse());

    if args.passkey.is_none() || args.username.is_none() {
        eprintln!("[ERROR] missing passkey or username, please provide them in the toml file or as CLI args");
        return;
    }

    download_torrent_files(args).await;
    #[cfg(feature = "sort")]
    sort_and_move();
}

async fn download_torrent_files(args: Config) {
    let passkey = args.passkey.unwrap();
    let username = args.username.unwrap();
    let search_type = args.search_type.unwrap();
    let starting_page = args.starting_page.unwrap();
    let ending_page = args.ending_page.unwrap();
    let max_size = args
        .max_size
        .unwrap_or("0".to_string())
        .parse::<ByteSize>()
        .unwrap_or(ByteSize::b(0))
        .as_u64();

    if starting_page > ending_page {
        eprintln!("[ERROR] starting page was greater than ending page");
        return;
    }

    if search_type != "anime" && search_type != "music" {
        eprintln!("[ERROR] invalid search_type, please select `anime` or `music`");
        return;
    }

    let filename_regex = Regex::new(r"Filename=.(?P<filename>.*).").unwrap();
    let sleep_time = Duration::from_millis(100);

    let client = HTTPClient::new();
    for page in starting_page..ending_page + 1 {
        let response = match client
            .get("https://animebytes.tv/scrape.php")
            .query(&[
                ("torrent_pass", &passkey),
                ("username", &username),
                ("type", &search_type),
                ("page", &page.to_string()),
                ("sort", &"size".to_string()),
                ("way", &"asc".to_string()),
                ("limit", &"50".to_string()),
            ])
            .send()
            .await
        {
            Ok(resp) => resp,
            Err(err) => {
                eprintln!(
                    "[ERROR] could not make request to scrape endpoint, err={}",
                    err
                );
                return;
            }
        };

        let body = response.text().await.unwrap();

        // i could have handled error messages with the same ScrapeResult struct but this was easy /shrug
        match serde_json::from_str::<ABError>(&body) {
            Ok(ab_err) => {
                eprintln!("[ERROR] {}", ab_err.error);
                return;
            }
            Err(_) => (),
        }

        let groups = match serde_json::from_str::<ScrapeResult>(&body) {
            Ok(anime_result) => anime_result.groups,
            Err(err) => {
                eprintln!("[ERROR] could not read response - {:#?}", err);
                write("./errored_on_response.json", body).unwrap();
                return;
            }
        };

        for release in groups.into_iter() {
            for torrent in release.torrents.into_iter() {
                if max_size > 0 && torrent.size > max_size {
                    println!(
                        "[INFO] skipping https://animebytes.tv/torrents.php?id={}&torrentid={} due to size",
                        release.id, torrent.id
                    );
                    sleep(sleep_time).await;
                    continue;
                }

                let dl_link = torrent.link.to_owned();
                let response = client.get(dl_link).send().await.unwrap();
                let headers = response.headers().to_owned();

                // content-disposition header contains the filename you would get if you had just hit the DL button on the site
                // i'm only going through the effort of grabbing it because i use the filename to sort the torrent files afterwards
                let content_disposition_header = headers.get(header::CONTENT_DISPOSITION).unwrap();

                // the to_str trait for HeaderValue treats the header as ascii which doesn't work for a lot of content on AB
                // instead we'll read the bytes and convert them to a UTF-8 string
                let content_disposition_bytes = content_disposition_header.as_bytes();
                let content_disposition = match from_utf8(content_disposition_bytes) {
                    Ok(val) => val,
                    Err(err) => {
                        println!("{:#?}", content_disposition_header);
                        println!("[ERROR] could not convert header - {:#?}", err);
                        continue;
                    }
                };

                let data = response.bytes().await.unwrap();

                let torrent_file_name = filename_regex
                    .captures(content_disposition)
                    .unwrap()
                    .name("filename")
                    .unwrap()
                    .as_str()
                    .trim_start_matches("."); // hidden files get ignored by some torrent clients and some file names will start with a period

                write(torrent_file_name, data).unwrap();

                // this will ensure we don't hit the scrape endpoint sooner than requested by the api docs (3s)
                // each page should take a bit over 5 seconds assuming 1 torrent per group
                sleep(sleep_time).await;
            }
        }
        println!("[INFO] page {}/{} complete", page, &ending_page)
    }
}

#[test]
fn test_ab_response_parsing() {
    let response = std::fs::read_to_string("./ab_response.json").unwrap();

    match serde_json::from_str::<ScrapeResult>(&response) {
        Ok(scrape_result) => {
            let group = scrape_result.groups.first().unwrap();
            let torrent = group.torrents.first().unwrap();
            assert_eq!(
                torrent.link,
                "https://animebytes.tv/torrent/45794/download/"
            );
            assert_eq!(torrent.size, 2656532);
        }
        Err(err) => panic!("[ERROR] could not parse response {:#?}", err),
    };
}

#[test]
fn test_content_disposition_regex() {
    let content_disposition =
        "attachment; Filename=\"What is Fate... - Oneshot [2002] - Translated.36645.torrent\"";
    let re = Regex::new(r"Filename=.(?P<filename>.*).").unwrap();
    let captures = re.captures(content_disposition);
    let torrent_file_name = captures.unwrap().name("filename").unwrap().as_str();
    assert_eq!(
        torrent_file_name,
        "What is Fate... - Oneshot [2002] - Translated.36645.torrent"
    );
}
